
/****************************************************************************/
/*                                                                          */
/* cmult.c - Multiple Connection "ftp" code.                                */
/*                                                                          */
/* Mark Allman (mallman@cs.ohiou.edu)                                       */
/* Ohio University Computer Science                                         */
/*                                                                          */
/* Copyright (c) 1996 Mark Allman                                           */
/*                                                                          */
/****************************************************************************/

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/types.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <arpa/inet.h>
#include <arpa/ftp.h>
#include <arpa/telnet.h>

#include <ctype.h>
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <varargs.h>

#include "ftp_var.h"

extern int h_errno;
extern jmp_buf	recvabort;

extern struct	sockaddr_in hisctladdr;
static struct  	sockaddr_in data_addr [MAXCNNCTS];
int      	mdata [MAXCNNCTS], flag = 0;
extern int	abrtflag, ptabflg, ptflag;
extern jmp_buf	ptabort;
extern struct	sockaddr_in myctladdr;
extern off_t	restart_point;
char            mprtcmd [CMDLEN];
struct cnninfo  conn [MAXCNNCTS];
unsigned int cnncts = 1, servcnn = 0, cnnuse [MAXCNNCTS], xtrawrite = 0;
int useful = 0;
unsigned long sbufsize = DEFSOCKBUFSIZE;
unsigned int stock = 1, mstats = 1, rttoff [MAXCNNCTS];
static jmp_buf	sendabort;
extern FILE *gf;
sig_t oldintr, oldintp;
int (*closefunc) __P((FILE *));
int aflag = FALSE;

#define HASHBYTES 1024

int minitconn ();
int gettimeofday ();

/****************************************************************************/
/*                                                                          */
/* Toggle whether or not to show all the ugly XFTP statistics.              */
/*                                                                          */
/****************************************************************************/

void stats ()
{
    if (mstats == TRUE)
    {
	mstats = FALSE;
	fprintf (stdout,"Statistics will not be displayed.\n");
    }
    else
    {
	mstats = TRUE;
	fprintf (stdout,"Statistics will be displayed.\n");
    }
}


/****************************************************************************/
/*                                                                          */
/* Simple user version of the "NOOP" command.  For testing.                 */
/*                                                                          */
/****************************************************************************/

void echo ()
{
    unsigned short status;

    status = command ("NOOP");
}

/****************************************************************************/
/*                                                                          */
/* Toggle between using "stock" code and "OU" code for one connection       */
/* transfers.                                                               */
/*                                                                          */
/****************************************************************************/

void setstock ()
{
    unsigned short status;

    if (stock == FALSE)
    {
	if ((status = command ("STCK 1")) == COMPLETE)
	    stock = TRUE;
	else
	    fprintf (stderr,"STCK command failed.\n");
    }
    else
    {
	if ((status = command ("STCK 0")) == COMPLETE)
	    stock = FALSE;
	else
	    fprintf (stderr,"STCK command failed.\n");
    }
}

/****************************************************************************/
/*                                                                          */
/* Use the "MULT" command to setup the number of TCP connections to use.    */
/*                                                                          */
/****************************************************************************/

void setmult (argc,argv)
int argc;
char *argv [];
{
    unsigned int tmp;
    unsigned short status;

    if (argc > 1)
	tmp = atoi (argv [1]);
    else
	tmp = DEFCNNCTS;
    if (tmp <= 0 || tmp > MAXCNNCTS)
	fprintf (stderr,"Invalid number of data connections.\n");
    else 
    {
	if (!servcnn)
	    if ((status = command ("MULT")) == COMPLETE)
		sscanf (reply_string, "200 %d", &servcnn);
	if (servcnn || (status == COMPLETE))
	{
	    if (tmp <= servcnn)
		cnncts = tmp;
	    else
		cnncts = servcnn;
	}
	fprintf (stdout,"Using %d data connection(s).\n", cnncts);
    }
}

/****************************************************************************/
/*                                                                          */
/* Use the "WIN" command to set the TCP send/receive window size.           */
/*                                                                          */
/****************************************************************************/

void setwin (argc,argv)
int argc;
char *argv [];
{
    unsigned int tmp;
    unsigned short status;

    if (argc > 1)
	tmp = atoi (argv [1]);
    else
	tmp = DEFSOCKBUFSIZE;
    if ((tmp * 1024) <= 0 || (tmp * 1024) > MAXBUFSIZE)
	fprintf (stderr,"Invalid window size: %d\n", tmp);
    else
    {
	status = command ("WIN %d", tmp);
	if (status == COMPLETE)
	    sbufsize = tmp * 1024;
    }
}

/****************************************************************************/
/*                                                                          */
/* Accept a data connection.                                                */
/*                                                                          */
/****************************************************************************/

int acceptcnn(i,lmode,from,fromlen)
int i;
char *lmode;
struct sockaddr_in *from;
int *fromlen;
{
    int s, tos;

    s = accept(mdata [i], (struct sockaddr *) from, fromlen);
    if (s < 0)
    {
	if (errno != EWOULDBLOCK)
	{
	    warn("accept");
	    (void) close(mdata [i]), mdata [i] = -1;
	    return (-1);
	}
	return (1);
    }
    if (hash)
    {
	(void) putchar('*');
	fflush (stdout);
    }
    useful++;
    (void) close(mdata [i]);
    mdata [i] = s;
    
#ifdef IP_TOS
    tos = IPTOS_THROUGHPUT;
    if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int)) < 0)
	warn("setsockopt TOS (ignored)");
#endif

    if (fcntl (mdata [i],F_SETFL,FNDELAY) < 0)
	warn ("fcntl (non-blocking)");
    return (0);
}

/****************************************************************************/
/*                                                                          */
/* Write at most one chunk of data to the given data connection.            */
/*                                                                          */
/****************************************************************************/

int writecnn (fin,i,noread,curoff,bytes)
FILE *fin;
int i, *noread, *curoff, *bytes;
{
    int bw;

    if (conn [i].size == 0 && !(*noread))
    {
	if ((conn [i].size = 
	     read (fileno (fin), conn [i].buf.data,MAXDATA)) < 0)
	{
	    warn ("read () failed");
	    return (-1);
	}
	else if (conn [i].size == 0)
	{
	    *noread = 1;
	    conn [i].size = -1;
	    return (0);
	}
	conn [i].pos = 0;
	conn [i].size += sizeof (unsigned long);
	conn [i].buf.offset = *curoff;
	*curoff += (conn [i].size - 4);
	*bytes += (conn [i].size - 4);
	cnnuse [i]++;
    }
    bw = write (mdata [i],((char *)&conn [i].buf + conn [i].pos), 
		conn [i].size);
    if (bw < 0)
    {
	if (errno == EWOULDBLOCK)
	    return (0);
	else
	{
	    warn ("write (from conn)");
	    return (-1);
	}
    }
    if (bw != conn [i].size)
    {
	conn [i].size -= bw;
	conn [i].pos += bw;
	return (0);
    }
    if (*noread)
    {
	conn [i].size = -1;
	return (0);
    }
    conn [i].size = 0;
    return (1);
}


/****************************************************************************/
/*                                                                          */
/* Send a file to the server using one or more connections.                 */
/*                                                                          */
/****************************************************************************/

void msendrequest(cmd, local, remote, printnames)
char *cmd, *local, *remote;
int printnames;
{
    struct stat st;
    struct timeval start1, stop1;

    int d, i, j = -1, astat;
    FILE *fin, *popen();
    sig_t oldintr, oldintp;
    long bytes = 0, hashbytes = HASHBYTES;
    char *lmode;
    struct fd_set wmask, rmask;
    int result, blocked = -1;
    int accflag = FALSE;
    unsigned short maxfd = 0;
    struct sockaddr_in from;
    int s, fromlen = sizeof (from), tos, noread, curoff = 0;
    int alldone;
    int onlyone = 6, writing = 0, ok = 0, notok = 0;

    if (proxy)
    {
	fprintf (stderr,"Multiple connection proxy transfers not supported.\n");
	return;
    }
    useful = 0;
    if (verbose && printnames) 
    {
	if (local && *local != '-')
	    printf("local: %s ", local);
	if (remote)
	    printf("remote: %s\n", remote);
    }
    if (proxy) 
    {
	proxtrans(cmd, local, remote);
	return;
    }
    if (curtype != type)
	changetype(type, 0);
    closefunc = NULL;
    oldintr = NULL;
    oldintp = NULL;
    lmode = (char *)malloc (sizeof (char) * 5);
    strcpy (lmode,"w");

    if (setjmp(sendabort)) 
    {
	while (cpend)                      /* if != 0, server reply pending */
	    (void) getreply(0);
	if (data >= 0)                     /* if data connection is open ?? */
	{
	    (void) close(data);
	    data = -1;
	}
	if (oldintr)
	    (void) signal(SIGINT,oldintr);
	if (oldintp)
	    (void) signal(SIGPIPE,oldintp);
	code = -1;                         /* return/reply code for command */
	return;
    }

    oldintr = signal(SIGINT, abortsend);
    if (strcmp(local, "-") == 0)           /* wow!  file from stdin??? */
	fin = stdin;
    else if (*local == '|')                /* if reading from a pipe */
    {
	oldintp = signal(SIGPIPE,SIG_IGN);
	fin = popen(local + 1, "r");
	if (fin == NULL) 
	{
	    warn("%s", local + 1);
	    (void) signal(SIGINT, oldintr);
	    (void) signal(SIGPIPE, oldintp);
	    code = -1;
	    return;
	}
	closefunc = pclose;
    }
    else 
    {
	fin = fopen(local, "r");           /* open the local file */
	if (fin == NULL) 
	{
	    warn("local: %s", local);
	    (void) signal(SIGINT, oldintr);
	    code = -1;
	    return;
	}
	closefunc = fclose;
	if (fstat(fileno(fin), &st) < 0 ||
	    (st.st_mode&S_IFMT) != S_IFREG) 
	{
	    fprintf(stdout, "%s: not a plain file.\n", local);
	    (void) signal(SIGINT, oldintr);
	    fclose(fin);
	    code = -1;
	    return;
	}
    }

    for (i = 0; i < MAXCNNCTS; i++)
    {
	mdata [i] = -1;
	conn [i].size = 0;
    }

    for (i = 0; i < cnncts; i++)
    {
	if (i > 0)
	    strcat (mprtcmd,",");
	if (minitconn(i))                    /* open TCP port for listening */
	{
	    (void) signal(SIGINT, oldintr);
	    if (oldintp)
		(void) signal(SIGPIPE, oldintp);
	    code = -1;
	    if (closefunc != NULL)
		(*closefunc)(fin);
	    return;
	}
	if (setsockopt (mdata [i],SOL_SOCKET,SO_SNDBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    warn ("setsockopt - ftp - send");
	if (setsockopt (mdata [i],SOL_SOCKET,SO_RCVBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    warn ("setsockopt - ftp - recv");    
    }

    result = command (mprtcmd);              /* the constructed command */
    if (result == ERROR && sendport == -1) 
	sendport = 0;

    if (setjmp(sendabort))
	goto abort;

    if (restart_point &&
	(strcmp(cmd, "STOR") == 0 || strcmp(cmd, "APPE") == 0)) 
    {
	int rc;
	
	switch (curtype) 
	{
	    case TYPE_A:
	    rc = fseek(fin, (long) restart_point, SEEK_SET);
	    break;

	    case TYPE_I:
	    case TYPE_L:
	    rc = lseek(fileno(fin), restart_point, SEEK_SET);
	    break;
	}
	if (rc < 0) 
	{
	    warn("local: %s", local);
	    restart_point = 0;
	    if (closefunc != NULL)
		(*closefunc)(fin);
	    return;
	}
	if (command("REST %ld", (long) restart_point) != CONTINUE) 
	{
	    restart_point = 0;
	    if (closefunc != NULL)
		(*closefunc)(fin);
	    return;
	}
	restart_point = 0;
	strcpy (lmode,"r+w");
    }
    if (remote) 
    {
	if (command("%s %s", cmd, remote) != PRELIM) 
	{
	    (void) signal(SIGINT, oldintr);
	    if (oldintp)
		(void) signal(SIGPIPE, oldintp);
	    if (closefunc != NULL)
		(*closefunc)(fin);
	    return;
	}
    } 
    else
	if (command("%s", cmd) != PRELIM) 
	{
	    (void) signal(SIGINT, oldintr);
	    if (oldintp)
		(void) signal(SIGPIPE, oldintp);
	    if (closefunc != NULL)
		(*closefunc)(fin);
	    return;
	}

    for (i = 0; i < cnncts; i++)
    {    
	if (fcntl (mdata [i],F_SETFL,FNDELAY) < 0)
	    warn ("fcntl (non-blocking)");
	s = accept(mdata [i], (struct sockaddr *) &from, &fromlen);
	if (s < 0)
	{
	    if (errno == EWOULDBLOCK) 
		conn [i].accptd = FALSE;
	    else
	    {
		warn("accept");
		(void) close(mdata [i]), mdata [i] = -1;
		goto abort;
	    }
	}
	else
	{
	    if (hash)
	    {
		(void) putchar('+');     /* new hash marks for connection */
		fflush (stdout);         /* acceptance */
	    }
	    conn [i].accptd = TRUE;
	    accflag = TRUE;
	    (void) close(mdata [i]);
	    mdata [i] = s;

#ifdef IP_TOS
	    tos = IPTOS_THROUGHPUT;
	    if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos, 
			   sizeof(int)) < 0)
		warn("setsockopt TOS (ignored)");
#endif
	    if (fcntl (mdata [i],F_SETFL,FNDELAY) < 0)
		warn ("fcntl (non-blocking)");
	    cnnuse [i] = 0;
	    useful++;
	}
    }

    oldintp = signal(SIGPIPE, SIG_IGN);
    (void) gettimeofday(&start1, (struct timezone *)0);

    switch (curtype) 
    {
	case TYPE_I:
	case TYPE_L:
	    errno = d = noread = 0;
	    while (1)
	    {
		writing = 0;
		FD_ZERO (&wmask);
		FD_ZERO (&rmask);

		alldone = TRUE;
		for (i = 0; i < cnncts; i++)
		    if (!conn [i].accptd)
		    {
			FD_SET (mdata [i],&rmask);
			conn [i].ready = FALSE;
			if (mdata [i] > maxfd)
			    maxfd = mdata [i];
			alldone = FALSE;
		    }
		    else if (conn [i].size >= 0)
		    {
			FD_SET (mdata [i],&wmask);
			conn [i].ready = TRUE;
			if (mdata [i] > maxfd)
			    maxfd = mdata [i];
			alldone = FALSE;
		    }
		    else
			conn [i].ready = FALSE;
		if (alldone && noread)
		    break;

		if ((j = select (maxfd + 1,&rmask,&wmask,0,0)) < 0)
		    warn ("select (main)");
		blocked = 0;
		for (i = 0; i < cnncts; i++)
		    if (!conn [i].accptd && FD_ISSET (mdata [i],&rmask))
		    {
			if ((astat = acceptcnn (i,lmode,&from,&fromlen)) < 0)
			    goto abort;
			if (astat > 0)
			    continue;
			conn [i].accptd = TRUE;
			accflag = TRUE;
			cnnuse [i] = 0;
			blocked++;
			conn [i].ready = FALSE;
		    }
		while (blocked < j && j > 0)
		{
		    for (i = 0; i < cnncts; i++)
			if (conn [i].ready && conn [i].size >= 0 &&
			    FD_ISSET (mdata [i],&wmask))
			{
			    writing++;
			    if ((conn [i].ready = writecnn (fin,i,&noread,
							    &curoff,&bytes)) 
				< 0)
				goto abort;
			    if (conn [i].ready == 0)
				blocked++;
			    if (conn [i].size == -1)
				useful--;

			}
		        else if (conn [i].size == 0 && noread)
			{
			    conn [i].size = -1;
			    useful--;
			}
		}
		if (hash) 
		{
		    while (bytes >= hashbytes) 
		    {
			(void) putchar('#');
			hashbytes += HASHBYTES;
		    }
		    (void) fflush(stdout);
		}
		if (accflag && useful <= 0)
		    break;
	    }
	    if (hash && bytes > 0) 
	    {
		if (bytes < HASHBYTES)
		    (void) putchar('#');
		(void) putchar('\n');
		(void) fflush(stdout);
	    }
	    break;

	case TYPE_A:
    	    printf ("ASCII transfers not yet supported.\n");
	    break;
    }
    if (closefunc != NULL && fin != NULL)
	(*closefunc)(fin);
    if (gf != NULL)
	fflush (gf);
    for (i = 0; i < cnncts; i++)
    {
	if (mdata [i] >= 0 && (close (mdata [i]) < 0))
	    warn ("close");
	mdata [i] = -1;
    }
    fflush (stdout);

    (void) getreply(0);
    (void) gettimeofday(&stop1, (struct timezone *)0);
    (void) signal(SIGINT, oldintr);
    if (oldintp)
	(void) signal(SIGPIPE, oldintp);
    if (bytes > 0)
    {
	ptransfer("sent", bytes, &start1, &stop1);
	if (mstats)
	{
	    fprintf (stdout,"Connection usage:\n");
	    for (i = 0; i < cnncts; i++)
		fprintf (stdout,"\tconnection [%d] == %d\n", i, cnnuse [i]);
	}
    }
    return;
abort:
    (void) gettimeofday(&stop1, (struct timezone *)0);
    (void) signal(SIGINT, oldintr);
    if (oldintp)
	(void) signal(SIGPIPE, oldintp);
    if (!cpend) 
    {
	code = -1;
	return;
    }
    for (i = 0; i < cnncts; i++)
    {
	if (mdata [i] >= 0 && (close (mdata [i]) < 0))
	    warn ("close");
	mdata [i] = -1;
    }
    (void) getreply(0);
    code = -1;
    if (closefunc != NULL && fin != NULL)
	(*closefunc)(fin);
    if (bytes > 0)
    {
	ptransfer("sent", bytes, &start1, &stop1);
	if (mstats)
	{
	    fprintf (stdout,"Connection usage:\n");
	    for (i = 0; i < cnncts; i++)
		fprintf (stdout,"\tconnection [%d] == %d\n", i, cnnuse [i]);
	}
    }
}


/****************************************************************************/
/*                                                                          */
/* Start listening on the data connection(s).                               */
/*                                                                          */
/****************************************************************************/

int minitconn(n)
int n;
{
    char *p, *a;
    int len, tmpno = 0;
    int on = 1;
    char tmp [CMDLEN];

    data_addr [n] = myctladdr;
    if (sendport)
	data_addr [n].sin_port = 0;	/* let system pick one */ 
    if (mdata [n] != -1)
	(void) close(mdata [n]);
    mdata [n] = socket(AF_INET, SOCK_STREAM, 0);
    if (mdata [n] < 0) 
    {
	warn("socket");
	if (tmpno)
	    sendport = 1;
	return (1);
    }
    if (!sendport)
	if (setsockopt(mdata [n], SOL_SOCKET, SO_REUSEADDR, (char *)&on, 
		       sizeof (on)) < 0) 
	{
	    warn("setsockopt (reuse address)");
	    goto bad;
	}
    if (bind(mdata [n], (struct sockaddr *)&data_addr [n], sizeof 
	     (data_addr [n])) < 0) 
    {
	warn("bind");
	goto bad;
    }
    if (options & SO_DEBUG &&
	setsockopt(mdata [n], SOL_SOCKET, SO_DEBUG, (char *)&on, sizeof (on)) 
	< 0)
	warn("setsockopt (ignored)");
    len = sizeof (data_addr [n]);
    if (getsockname(mdata [n], (struct sockaddr *)&data_addr [n], &len) < 0) 
    {
	warn("getsockname");
	goto bad;
    }
    if (listen(mdata [n], 1) < 0)
	warn("listen");
    if (sendport) 
    {
#define	UC(b)	(((int)b)&0xff)
	if (n == 0)
	{	    
	    a = (char *)&data_addr [n].sin_addr;
	    sprintf (mprtcmd,"MPRT %d,%d,%d,%d,",UC(a[0]), UC(a[1]), UC(a[2]), 
		     UC(a[3]));
	}
	p = (char *)&data_addr [n].sin_port;
	sprintf (tmp,"%d,%d",UC(p[0]), UC(p[1]));
	strcat (mprtcmd,tmp);
	return (0);
    }
    if (tmpno)
	sendport = 1;
#ifdef IP_TOS
    on = IPTOS_THROUGHPUT;
    if (setsockopt(mdata [n], IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
	warn("setsockopt TOS (ignored)");
#endif
    return (0);
bad:
    (void) close(mdata [n]), mdata [n] = -1;
    if (tmpno)
	sendport = 1;
    return (1);
}

/****************************************************************************/
/*                                                                          */
/* Receive a file using multiple data connections.                          */
/*                                                                          */
/****************************************************************************/

void mrecvrequest(cmd, local, remote, lmode, printnames)
char *cmd, *local, *remote, *lmode;
int printnames;
{
    FILE *fout, *din = 0;
    int (*closefunc) __P((FILE *));
    sig_t oldintr, oldintp;
    int c, d, is_retr, tcrflag, bare_lfs = 0;
    static int bufsize;
    static char *buf;
    long bytes = 0, hashbytes = HASHBYTES;
    struct timeval start, stop;
    struct stat st;
    char *tmp;
    struct sockaddr_in from;
    int s, fromlen = sizeof (from), i, j, maxfd, rb;
    int dflag [MAXCNNCTS], result, tos, doneflag, bw;
    struct fd_set rmask;
    struct holding hold [MAXCNNCTS];
    struct mdata mbuf;

    if (proxy)
    {
	fprintf (stderr,"Multiple connection proxy transfers not supported.\n");
	return;
    }
    aflag = FALSE;
    is_retr = strcmp(cmd, "RETR") == 0;
    if (is_retr && verbose && printnames) 
    {
	if (local && *local != '-')
	    printf("local: %s ", local);
	if (remote)
	    printf("remote: %s\n", remote);
    }
    if (proxy && is_retr) 
    {
	proxtrans(cmd, local, remote);
	return;
    }
    closefunc = NULL;
    oldintr = NULL;
    oldintp = NULL;
    tcrflag = !crflag && is_retr;
    if (setjmp(recvabort)) 
    {
	while (cpend) 
	    (void) getreply(0);
	if (data >= 0) 
	{
	    (void) close(data);
	    data = -1;
	}
	if (oldintr)
	    (void) signal(SIGINT, oldintr);
	code = -1;
	return;
    }
    oldintr = signal(SIGINT, abortrecv);
    if (strcmp(local, "-") && *local != '|') 
    {
	if (access(local, 2) < 0) {
	    char *dir = strrchr(local, '/');

	    if (errno != ENOENT && errno != EACCES) 
	    {
		warn("local: %s", local);
		(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	    }
	    if (dir != NULL)
		*dir = 0;
	    d = access(dir ? local : ".", 2);
	    if (dir != NULL)
		*dir = '/';
	    if (d < 0) {
		warn("local: %s", local);
		(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	    }
	    if (!runique && errno == EACCES &&
		chmod(local, 0600) < 0) 
	    {
		warn("local: %s", local);
		(void) signal(SIGINT, oldintr);
		(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	    }
	    if (runique && errno == EACCES &&
		(local = gunique(local)) == NULL) 
	    {
		(void) signal(SIGINT, oldintr);
		code = -1;
		return;
	    }
	}
	else if (runique && (local = gunique(local)) == NULL) 
	{
	    (void) signal(SIGINT, oldintr);
	    code = -1;
	    return;
	}
    }
    if (!is_retr) 
    {
	if (curtype != TYPE_A)
	    changetype(TYPE_A, 0);
    } 
    else if (curtype != type)
	changetype(type, 0);

    for (i = 0; i < MAXCNNCTS; i++)
    {
	mdata [i] = -1;
	conn [i].size = 0;
    }
    for (i = 0; i < cnncts; i++)
    {
	if (i > 0)
	    strcat (mprtcmd,",");
	if (minitconn(i))                    /* open TCP port for listening */
	{
	    (void) signal(SIGINT, oldintr);
	    if (oldintp)
		(void) signal(SIGPIPE, oldintp);
	    code = -1;
	    return;
	}
	if (setsockopt (mdata [i],SOL_SOCKET,SO_SNDBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    warn ("setsockopt - ftp - send");
	if (setsockopt (mdata [i],SOL_SOCKET,SO_RCVBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    warn ("setsockopt - ftp - recv");    
    }

    fprintf (stdout,"mprtcmd == %s\n", mprtcmd);
    result = command (mprtcmd);              /* the constructed command */
    if (result == ERROR)
	goto abort;

    if (setjmp(recvabort))
	goto abort;
    if (is_retr && restart_point &&
	command("REST %ld", (long) restart_point) != CONTINUE)
	return;
    if (remote) 
    {
	if ((result = command("%s %s", cmd, remote)) != PRELIM) 
	{
	    (void) signal(SIGINT, oldintr);
	    return;
	}
    } 
    else 
    {
	if (command("%s", cmd) != PRELIM) 
	{
	    (void) signal(SIGINT, oldintr);
	    return;
	}
    }

    for (i = 0; i < cnncts; i++)
    {    
	if (fcntl (mdata [i],F_SETFL,FNDELAY) < 0)
	    warn ("fcntl (non-blocking)");
	s = accept(mdata [i], (struct sockaddr *) &from, &fromlen);
	if (s < 0)
	{
	    if (errno == EWOULDBLOCK) 
		conn [i].accptd = FALSE;
	    else
	    {
		warn("accept");
		(void) close(mdata [i]), mdata [i] = -1;
		goto abort;
	    }
	}
	else
	{
	    if (hash)
	    {
		(void) putchar('+');     /* new hash marks for connection */
		fflush (stdout);         /* acceptance */
	    }
	    conn [i].accptd = TRUE;
	    aflag = TRUE;
	    (void) close(mdata [i]);
	    mdata [i] = s;

#ifdef IP_TOS
	    tos = IPTOS_THROUGHPUT;
	    if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos, 
			   sizeof(int)) < 0)
		warn("setsockopt TOS (ignored)");
#endif
	    if (fcntl (mdata [i],F_SETFL,FNDELAY) < 0)
		warn ("fcntl (non-blocking)");
	    cnnuse [i] = 0;
	    useful++;
	}
    }

    if (strcmp(local, "-") == 0)
	fout = stdout;
    else if (*local == '|') 
    {
	oldintp = signal(SIGPIPE, SIG_IGN);
	fout = popen(local + 1, "w");
	if (fout == NULL) 
	{
	    warn("%s", local+1);
	    goto abort;
	}
	closefunc = pclose;
    } 
    else 
    {
	fout = fopen(local, lmode);
	if (fout == NULL) 
	{
	    warn("local: %s", local);
	    goto abort;
	}
	closefunc = fclose;
    }
    if (fstat(fileno(fout), &st) < 0 || st.st_blksize == 0)
	st.st_blksize = BUFSIZ;
    if (st.st_blksize > bufsize) 
    {
	if (buf)
	    (void) free(buf);
	buf = malloc((unsigned)st.st_blksize);
	if (buf == NULL) 
	{
	    warn("malloc");
	    bufsize = 0;
	    goto abort;
	}
	bufsize = st.st_blksize;
    }
    (void) gettimeofday(&start, (struct timezone *)0);
    switch (curtype) 
    {

	case TYPE_I:
	case TYPE_L:
	    for (i = 0; i < cnncts; i++)
	    {
		if (mdata [i] > maxfd)
		    maxfd = mdata [i];
		if (conn [i].accptd)
		    dflag [i] = TRUE;
		else
		    dflag [i] = FALSE;
		hold [i].size = 0;
	    }
            while (1)
	    {
		doneflag = TRUE;
		FD_ZERO (&rmask);
		for (i = 0; i < cnncts; i++)
		{
		    FD_SET (mdata [i],&rmask);
		    if (dflag [i])
			doneflag = FALSE;
		}
		if (doneflag && aflag)
		    break;
		if ((j = select (maxfd + 1,&rmask,0,0,0)) < 0)
		{
		    warn ("select");
		    return;
		}
		if (!j)
		    continue;
		for (i = 0; i < cnncts; i++)
		    if (FD_ISSET (mdata [i],&rmask) && !conn [i].accptd)
			acceptcnn (i,lmode,&from,&fromlen);
		    else if (!conn [i].accptd)
			continue;
		    else if (FD_ISSET (mdata [i],&rmask) && hold [i].size)
		    {
			rb = read (mdata [i],&mbuf,sizeof (mbuf) -
				   hold [i].size);
			if (rb < 0)
			{
			    warn ("read");
			    break;
			}
			else if (rb > 0)
			{
			    tmp = (char *)(&hold [i].data) + hold [i].size;
			    memcpy (tmp,&mbuf,rb);
			    hold [i].size += rb;
			}
			else
			    dflag [i] = FALSE;
			if (hold [i].size == sizeof (mbuf) || !rb)
			{
			    if (fseek (fout,hold [i].data.offset,
				       SEEK_SET) < 0)
			    {
				warn ("fseek");
				goto abort;
			    }
			    if ((bw = fwrite (hold [i].data.data,sizeof (char),
					     hold [i].size - 
					     sizeof (mbuf.offset),
					     fout)) <= 0)
			    {
				warn ("fwrite");
				goto abort;
			    }
			    else
			    {
				cnnuse [i]++;
				bytes += bw;
				if (hash) 
				{
				    while (bytes >= hashbytes) 
				    {
					(void) putchar('#');
					hashbytes += HASHBYTES;
				    }
				    (void) fflush(stdout);
				}
			    }
			    hold [i].size = 0;
			}
		    }
		    else if (FD_ISSET (mdata [i],&rmask))
		    {
			rb = read (mdata [i],&mbuf,sizeof (mbuf));
			if (rb < 0)
			{
			    warn ("read");
			    break;
			}
			if (rb == sizeof (mbuf))
			{
			    if (fseek (fout,mbuf.offset,SEEK_SET) < 0)
			    {
				warn ("fseek");
				goto abort;
			    }
			    if ((bw = fwrite (mbuf.data,sizeof (char),
					rb - sizeof (mbuf.offset),
					fout)) == 0)
			    {
				warn ("fwrite");
				goto abort;
			    }
			    else
			    {
				cnnuse [i]++;
				bytes += bw;
				if (hash) 
				{
				    while (bytes >= hashbytes) 
				    {
					(void) putchar('#');
					hashbytes += HASHBYTES;
				    }
				    (void) fflush(stdout);
				}
			    }
			    hold [i].size = 0;
			}
			else if (!rb)
			    dflag [i] = FALSE;
			else
			{
			    memcpy ((char *)(&hold [i].data),&mbuf,rb);
			    hold [i].size = rb;
			}
		    }
		if (i < cnncts)
		    break;
	    }
	    if (!doneflag && i < cnncts)
		return;
	    break;

	case TYPE_A:
	    fprintf (stderr,"Multiple connection ASCII transfers not supported.\n");
	    break;
    }

    if (closefunc != NULL)
	(*closefunc)(fout);
    (void) signal(SIGINT, oldintr);
    if (oldintp)
	(void) signal(SIGPIPE, oldintp);
    if (hash) 
    {
	while (bytes >= hashbytes) 
	{
	    (void) putchar('#');
	    hashbytes += HASHBYTES;
	}
	(void) putchar ('\n');
	(void) fflush (stdout);
    }
    (void) getreply(0);
    (void) gettimeofday(&stop, (struct timezone *)0);
    if (bytes > 0 && is_retr)
    {
	ptransfer("received", bytes, &start, &stop);
	if (mstats)
	{
	    fprintf (stdout,"Connection usage:\n");
	    for (i = 0; i < cnncts; i++)
		fprintf (stdout,"\tconnection [%d] == %d\n", i, cnnuse [i]);
	}
    }	
    for (i = 0; i < cnncts; i++)
    {
	fprintf (stdout,"mdata [%d] == %d\n", i, mdata [i]);
	if (mdata [i] >= 0 && (close (mdata [i]) < 0))
	    warn ("close");
	mdata [i] = -1;
    }
    return;
abort:

/* abort using RFC959 recommended IP,SYNC sequence  */

    (void) gettimeofday(&stop, (struct timezone *)0);
    if (oldintp)
	(void) signal(SIGPIPE, oldintr);
    (void) signal(SIGINT, SIG_IGN);
    if (!cpend) 
    {
	code = -1;
	(void) signal(SIGINT, oldintr);
	return;
    }

/* hmmmm... */
/*     abort_remote(din); */

    code = -1;
    if (closefunc != NULL && fout != NULL)
	(*closefunc)(fout);

    for (i = 0; i < cnncts; i++)
    {
	fprintf (stdout,"abort: mdata [%d] == %d\n", i, mdata [i]);
	if (mdata [i] >= 0 && (close (mdata [i]) < 0))
	    warn ("close");
	mdata [i] = -1;
    }
    if (bytes > 0)
    {
	ptransfer("received", bytes, &start, &stop);
	if (mstats)
	{
	    fprintf (stdout,"Connection usage:\n");
	    for (i = 0; i < cnncts; i++)
		fprintf (stdout,"\tconnection [%d] == %d\n", i, cnnuse [i]);
	}
    }
    (void) signal(SIGINT, oldintr);
}
