
/****************************************************************************/
/*                                                                          */
/* smult.c - Multiple Connection "ftp" code.                                */
/*                                                                          */
/* Mark Allman (mallman@cs.ohiou.edu)                                       */
/* Ohio University Computer Science                                         */
/*                                                                          */
/* Copyright (c) 1996 Mark Allman                                           */
/*                                                                          */
/****************************************************************************/

#include <sys/param.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>

#include <arpa/ftp.h>
#include <arpa/inet.h>
#include <arpa/telnet.h>

#include <ctype.h>
#ifdef SUNOS4
#include <dirent.h>
#endif
#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <glob.h>
#include <limits.h>
#include <netdb.h>
#include <pwd.h>
#include <setjmp.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>

#include "oumisc.h"
#include "pathnames.h"
#include "extern.h"

#if __STDC__
#include <stdarg.h>
#else
#include <varargs.h>
#endif

#ifdef VERBOSE
FILE *fp;
#endif

extern	off_t restart_point;
extern	char cbuf[];
extern char *__progname;

extern struct	sockaddr_in ctrl_addr;
extern struct	sockaddr_in data_source;
extern struct	sockaddr_in data_dest;
extern struct	sockaddr_in his_addr;
extern struct	sockaddr_in pasv_addr;
struct   sockaddr_in mdata_dest [MAXCNNCTS];   
struct   sockaddr_in mdata_source [MAXCNNCTS];
extern int	data;
int             mdata [MAXCNNCTS];             
extern jmp_buf	errcatch, urgcatch;
extern int	logged_in;
extern struct	passwd *pw;
extern int	debug;
extern int	timeout;    /* timeout after 15 minutes of inactivity */
extern int	maxtimeout; /* don't allow idle time to be set beyond 2 hours */
extern int	logging;
extern int	guest;
extern int	type;
extern int	form;
extern int	stru;			/* avoid C keyword */
extern int	mode;
extern int	usedefault;     	/* for data transfers */
extern int	pdata;      		/* for passive mode */
extern int      current;          
extern sig_atomic_t transflag;
extern off_t	file_size;
extern off_t	byte_count;
#if !defined(CMASK) || CMASK == 0
#undef CMASK
#define CMASK 027
#endif
extern int	defumask;         		/* default umask value */
extern char	tmpline[7];
extern char	hostname[MAXHOSTNAMELEN];
extern char	remotehost[MAXHOSTNAMELEN];

/*
 * Timeout intervals for retrying connections
 * to hosts that don't accept PORT cmds.  This
 * is a kludge, but given the problems with TCP...
 */
#define	SWAITMAX	90	/* wait at most 90 seconds */
#define	SWAITINT	5	/* interval between retries */

extern int	swaitmax;
extern int	swaitint;

#ifdef SETPROCTITLE
extern char	**Argv;        		/* pointer to argument vector */
extern char	*LastArgv;       	/* end of argv */
extern char	proctitle[LINE_MAX];	/* initial part of title */
#endif SETPROCTITLE

#define LOGCMD(cmd, file) \
	if (logging > 1) \
	    syslog(LOG_INFO,"%s %s%s", cmd, \
		*(file) == '/' ? "" : curdir(), file);
#define LOGCMD2(cmd, file1, file2) \
	 if (logging > 1) \
	    syslog(LOG_INFO,"%s %s%s %s%s", cmd, \
		*(file1) == '/' ? "" : curdir(), file1, \
		*(file2) == '/' ? "" : curdir(), file2);
#define LOGBYTES(cmd, file, cnt) \
	if (logging > 1) { \
		if (cnt == (off_t)-1) \
		    syslog(LOG_INFO,"%s %s%s", cmd, \
			*(file) == '/' ? "" : curdir(), file); \
		else \
		    syslog(LOG_INFO, "%s %s%s = %qd bytes", \
			cmd, (*(file) == '/') ? "" : curdir(), file, cnt); \
	}

static void	 ack __P((char *));
static void	 myoob __P((int));
static int	 checkuser __P((char *));
static FILE	*mdataconn __P((char *, off_t, char *, unsigned int));
static void	 dolog __P((struct sockaddr_in *));
char	        *curdir __P((void));
static void	 end_login __P((void));
static FILE	*getdatasock __P((char *));
static FILE	*mgetdatasock __P((char *,unsigned int));
char	        *gunique __P((char *));
static void	 lostconn __P((int));
static int	 receive_data __P((FILE *, FILE *));
static int	 mreceive_data __P((FILE **, FILE *));
static struct passwd *sgetpwnam __P((char *));
static char	*sgetsave __P((char *));
unsigned int cnncts = 1, stock = 1, multchk = FALSE;
unsigned long sbufsize = DEFSOCKBUFSIZE;
unsigned int maxcnn = MAXCNNCTS;
struct cnninfo conn [MAXCNNCTS];
char mprtcmd [CMDLEN];
static struct sockaddr_in data_addr [MAXCNNCTS];


/****************************************************************************/
/*                                                                          */
/* Setup multiple data connections (processing the "MULT" command).         */
/*                                                                          */
/****************************************************************************/

void multsetup ()
{
    unsigned int tmp;

    multchk = TRUE;
    reply (200," %d data connection(s) available.", maxcnn);
}

/****************************************************************************/
/*                                                                          */
/* Setup TCP send/receive window size (processing the "WIN" command).       */
/*                                                                          */
/****************************************************************************/

void winsetup (wsize)
char *wsize;
{
    unsigned int tmp;

    if (wsize != NULL)
    {
	tmp = atoi (wsize);
	if (((tmp * 1024) <= MAXBUFSIZE) && (tmp > 0))
	{
	    sbufsize = tmp * 1024;
	    reply (200," Window size set to %d.", sbufsize);
	}
	else
	    reply (500, " Invalid window size.  Valid range: 1 .. %d\n",
		   (MAXBUFSIZE / 1024));
    }
}

/****************************************************************************/
/*                                                                          */
/* Process the "STCK" command.  Chnage to either "stock" mode or "OU" mode  */
/* for single connection transfers.                                         */
/*                                                                          */
/****************************************************************************/

void stocksetup (val)
char *val;
{
    unsigned int tmp;

    stock = atoi (val);
    if (stock == TRUE)
	reply (211," Stock mode on.");
    else
	reply (211," Stock mode off.");
}

/****************************************************************************/
/*                                                                          */
/* Receive a file via one or more data connections (process "STOR"          */
/* command).                                                                */
/*                                                                          */
/****************************************************************************/

void mstore(name, mode, unique)
char *name, *mode;
int unique;
{
    FILE *fout, *din [MAXCNNCTS];
    struct stat st;
    int (*closefunc) __P((FILE *));
    char sizebuf[32];
    off_t size;
    int i;

    for (i = 0; i < cnncts; i++)
	mdata [i] = -1;

    if (unique && stat(name, &st) == 0 && (name = gunique(name)) == NULL) 
    {
	LOGCMD(*mode == 'w' ? "put" : "append", name);
	return;
    }
    
    if (restart_point)
	mode = "r+";
    fout = fopen(name, mode);
    closefunc = fclose;
    if (fout == NULL) 
    {
	perror_reply(553, name);
	LOGCMD(*mode == 'w' ? "put" : "append", name);
	return;
    }
    byte_count = -1;
    if (restart_point) 
    {
	if (type == TYPE_A) 
	{
	    off_t i, n;
	    int c;
	    
	    n = restart_point;
	    i = 0;
	    while (i++ < n) 
	    {
		if ((c=getc(fout)) == EOF) 
		{
		    perror_reply(550, name);
		    goto done;
		}
		if (c == '\n')
		    i++;
	    }
	    /*
	     * We must do this seek to "current" position
	     * because we are changing from reading to
	     * writing.
	     */
	    if (fseek(fout, 0L, L_INCR) < 0) 
	    {
		perror_reply(550, name);
		goto done;
	    }
	} 
	else if (lseek(fileno(fout), restart_point, L_SET) < 0) 
	{
	    perror_reply(550, name);
	    goto done;
	}
    }

    for (i = 0; i < cnncts; i++)
    {
	mdata_source [i].sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	din [i] = mdataconn(name, (off_t)-1, "r",i);
	if (din [i] == NULL)
	    goto done;
    }

    size = (off_t) - 1;
    if (size != (off_t) -1)
	(void) sprintf(sizebuf, " (%qd bytes)", size);
    else
	(void) strcpy(sizebuf, "");
    reply(150, "Opening %s mode data connection for '%s'%s.",
	  type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);

    if (mreceive_data(din, fout) == 0) 
    {
	if (unique)
	    reply(226, "Transfer complete (unique file name:%s)", name);
	else
	    reply(226, "Transfer complete.");
    }

    for (i = 0; i < cnncts; i++)
    {
	fclose (din [i]);
	mdata [i] = -1;
    }

    data = -1;
    pdata = -1;
    current = 0;                         /* reset current data connection */
done:
    LOGBYTES(*mode == 'w' ? "put" : "append", name, byte_count);
    (*closefunc)(fout);
}

/****************************************************************************/
/*                                                                          */
/* Open the given data connection.                                          */
/*                                                                          */
/****************************************************************************/

static FILE *mdataconn(name, size, mode, conn)
char *name;
off_t size;
char *mode;
unsigned int conn;
{
    char sizebuf[32];
    FILE *file;
    int retry = 0, tos;
    int on = 1;
    file_size = size;
    byte_count = 0;

    if (size != (off_t) -1)
	(void) sprintf(sizebuf, " (%qd bytes)", size);
    else
	(void) strcpy(sizebuf, "");
    if (pdata >= 0) 
    {
	struct sockaddr_in from;
	int s, fromlen = sizeof(from);

	s = accept(pdata, (struct sockaddr *)&from, &fromlen);
	if (s < 0) 
	{
	    reply(425, "Can't open data connection.");
	    (void) close(pdata);
	    pdata = -1;
	    current = 0;
	    return (NULL);
	}
	(void) close(pdata);
	pdata = s;

#ifdef IP_TOS
	tos = IPTOS_LOWDELAY;
	(void) setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&tos, sizeof(int));
#endif
	
	reply(150, "Opening %s mode data connection for '%s'%s.",
	      type == TYPE_A ? "ASCII" : "BINARY", name, sizebuf);
	return (fdopen(pdata, mode));
    }
    if (mdata [conn] >= 0) 
    {
	reply(125, "Using existing data connection for '%s'%s.",
	      name, sizebuf);
	usedefault = 1;
	return (fdopen(mdata [conn], mode));
    }
    file = mgetdatasock(mode,conn);
    if (file == NULL) 
    {
	reply(425, "Can't create data socket (%s,%d): %s.",
	      inet_ntoa(mdata_source [conn].sin_addr),
	      ntohs(mdata_source [conn].sin_port), strerror(errno));
	return (NULL);
    }
    mdata [conn] = fileno(file);
    if (fcntl (mdata [conn],F_SETFL,FNDELAY) < 0)
    {
	(void) fclose(file);
	mdata [conn] = -1;
	return (NULL);
    }
    if (setsockopt(mdata [conn], SOL_SOCKET, SO_REUSEADDR, (char *) &on, 
		   sizeof(on)) < 0)
    {
	(void) fclose(file);
	mdata [conn] = -1;
	return (NULL);
    }
    while (connect(mdata [conn], (struct sockaddr *)&mdata_dest [conn],
		   sizeof(mdata_dest [conn])) < 0) 
    {
	if (errno == EINPROGRESS)
	    break;
	else if (errno == EADDRINUSE && retry < swaitmax) 
	{
	    sleep((unsigned) swaitint);
	    retry += swaitint;
	    continue;
	}
	perror_reply(425, "Can't build data connection");

	(void) fclose(file);
	mdata [conn] = -1;
	return (NULL);
    }
    return (file);
}

/****************************************************************************/
/*                                                                          */
/* Read data from multiple data connections and write it to file.           */
/*                                                                          */
/****************************************************************************/

static int mreceive_data(instr, outstr)
FILE *instr [], *outstr;
{
    int c, i, j, doneflag = FALSE, dflag [MAXCNNCTS], rb;
    int cnt, bare_lfs = 0;
    unsigned short maxfd = 0;
    char buf [BUFSIZ], *tmp;
    struct fd_set rmask;
    struct mdata mbuf;
    short fdfg [MAXCNNCTS];
    struct holding hold [MAXCNNCTS];
    struct timeval timeout;

    transflag++;
    if (setjmp(urgcatch)) 
    {
	transflag = 0;
	return (-1);
    }
    timeout.tv_sec = 600;
    switch (type) 
    {
	
	case TYPE_I:
	case TYPE_L:
	    for (i = 0; i < cnncts; i++)
	    {
		if (fileno (instr [i]) > maxfd)
		    maxfd = fileno (instr [i]);
		dflag [i] = TRUE;
		hold [i].size = 0;
	    }
            while (1)
	    {
		doneflag = TRUE;
		FD_ZERO (&rmask);
		for (i = 0; i < cnncts; i++)
		    if (dflag [i])
		    {
			FD_SET (fileno (instr [i]),&rmask);
			doneflag = FALSE;
		    }
		if (doneflag)
		    break;
		if ((j = select (maxfd + 1,&rmask,0,0,&timeout)) <= 0)
		{
		    transflag = 0;
		    return (-1);
		}
		if (!j)
		    continue;
		for (i = 0; i < cnncts; i++)
		    if (!dflag [i])
			continue;
		    else if (FD_ISSET (fileno (instr [i]),&rmask) && 
			     hold [i].size)
		    {
			rb = read (fileno (instr [i]),&mbuf,sizeof (mbuf) -
				   hold [i].size);
			if (rb < 0)
			    break;
			else if (rb > 0)
			{
			    tmp = (char *)(&hold [i].data) + hold [i].size;
			    memcpy (tmp,&mbuf,rb);
			    hold [i].size += rb;
			}
			else
			    dflag [i] = FALSE;
			if (hold [i].size == sizeof (mbuf) || !rb)
			{
			    if (fseek (outstr,hold [i].data.offset,
				       SEEK_SET) < 0)
				goto file_err;
			    if (fwrite (hold [i].data.data,sizeof (char),
					     hold [i].size - 
					     sizeof (mbuf.offset),
					     outstr) <= 0)
				goto file_err;
			    hold [i].size = 0;
			}
		    }
		    else if (FD_ISSET (fileno (instr [i]),&rmask))
		    {
			rb = read (fileno (instr [i]),&mbuf,
				   sizeof (mbuf));
			if (rb < 0)
			    break;
			if (rb == sizeof (mbuf))
			{
			    if (fseek (outstr,mbuf.offset,SEEK_SET) < 0)
				goto file_err;
			    if (fwrite (mbuf.data,sizeof (char),
					rb - sizeof (mbuf.offset),
					outstr) == 0)
				goto file_err;
			    hold [i].size = 0;
			}
			else if (!rb)
			    dflag [i] = FALSE;
			else
			{
			    memcpy ((char *)(&hold [i].data),&mbuf,rb);
			    hold [i].size = rb;
			}
		    }
		if (i < cnncts)
		    break;
	    }
	    if (!doneflag && i < cnncts)
	    {
		transflag = 0;
		return (-1);
	    }

	    transflag = 0;
	    return (0);

        case TYPE_E:
	case TYPE_A:
	    reply(553, "TYPE E not implemented.");
	    transflag = 0;
	    return (-1);

	default:
	    reply(550, "Unimplemented TYPE %d in mreceive_data", type);
	    transflag = 0;
	    return (-1);
       }
data_err:
    transflag = 0;
    perror_reply(426, "Data Connection");
    return (-1);

file_err:
    transflag = 0;
    perror_reply(452, "Error writing file");
    return (-1);
}

/****************************************************************************/
/*                                                                          */
/* Open a data connection socket.                                           */
/*                                                                          */
/****************************************************************************/

static FILE *mgetdatasock(mode,conn)
char *mode;
unsigned int conn;
{
    int on = 1, s = -1, t, tries;

    if (mdata [conn] >= 0)
	return (fdopen(mdata [conn], mode));
    (void) seteuid((uid_t)0);
    s = socket(AF_INET, SOCK_STREAM, 0);
    if (s < 0)
	goto bad;
    if (setsockopt (s,SOL_SOCKET,SO_SNDBUF,(char *)&sbufsize,
		    sizeof (sbufsize)) < 0)
	goto bad;
    if (setsockopt (s,SOL_SOCKET,SO_RCVBUF,(char *)&sbufsize,
		    sizeof (sbufsize)) < 0)
	goto bad;
    if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on)) < 0)
	goto bad;

    /* anchor socket to avoid multi-homing problems */
    mdata_source [conn].sin_family = AF_INET;
    mdata_source [conn].sin_addr = ctrl_addr.sin_addr;

    for (tries = 1; ; tries++) {
	if (bind(s, (struct sockaddr *)&mdata_source [conn], 
		 sizeof(mdata_source [conn])) >= 0)
	    break;
	if (errno != EADDRINUSE || tries > 10)
	    goto bad;
	sleep(tries);
    }
    (void) seteuid((uid_t)pw->pw_uid);
#ifdef IP_TOS
    on = IPTOS_THROUGHPUT;
    if (setsockopt(s, IPPROTO_IP, IP_TOS, (char *)&on, sizeof(int)) < 0)
	syslog(LOG_WARNING, "setsockopt (IP_TOS): %m");
#endif
    return (fdopen(s, mode));
bad:
    /* Return the real value of errno (close may change it) */
    t = errno;
    (void) seteuid((uid_t)pw->pw_uid);
    (void) close(s);
    errno = t;
    return (NULL);
}

/****************************************************************************/
/*                                                                          */
/* Write at most one buffer of data to the given data connection.           */
/*                                                                          */
/****************************************************************************/

int writecnn (fin,i,noread,curoff)
FILE *fin;
int i, *noread, *curoff;
{
    int bw;

    if (conn [i].size == 0 && !(*noread))
    {
	if ((conn [i].size = 
	     read (fileno (fin), conn [i].buf.data,MAXDATA)) < 0)
	    return (-1);
	else if (conn [i].size == 0)
	{
	    *noread = 1;
	    conn [i].size = -1;
	    return (0);
	}
	conn [i].pos = 0;
	conn [i].size += sizeof (unsigned long);
	conn [i].buf.offset = *curoff;
	*curoff += (conn [i].size - 4);
    }
    bw = write (mdata [i],((char *)&conn [i].buf + conn [i].pos), 
		conn [i].size);
    if (bw < 0)
    {
	if (errno == EWOULDBLOCK)
	    return (0);
	else
	    return (-1);
    }
    if (bw != conn [i].size)
    {
	conn [i].size -= bw;
	conn [i].pos += bw;
	return (0);
    }
    if (*noread)
    {
	conn [i].size = -1;
	return (0);
    }
    conn [i].size = 0;
    return (1);
}

/****************************************************************************/
/*                                                                          */
/* Send file using multiple data connections.                               */
/*                                                                          */
/****************************************************************************/

static void msend_data(instr, outstr, blksize)
FILE *instr, *outstr;
off_t blksize;
{
    int c, cnt, filefd, netfd, d, noread, alldone, i, maxfd, j, blocked;
    int curoff = 0;
    char *buf;
    struct fd_set wmask;

    transflag++;
    if (setjmp(urgcatch)) 
    {
	transflag = 0;
	return;
    }
    switch (type) 
    {
	case TYPE_A:
	  while ((c = getc(instr)) != EOF) 
	  {
	      byte_count++;
	      if (c == '\n') 
	      {
		  if (ferror(outstr))
		      goto data_err;
		  (void) putc('\r', outstr);
	      }
	      (void) putc(c, outstr);
	  }
	  fflush(outstr);
	  transflag = 0;
	  if (ferror(instr))
	      goto file_err;
	  if (ferror(outstr))
	      goto data_err;
	  reply(226, "Transfer complete.");
	  return;

	case TYPE_I:
	case TYPE_L:

	    errno = d = noread = 0;
	    while (1)
	    {
		FD_ZERO (&wmask);

		alldone = TRUE;
		for (i = 0; i < cnncts; i++)
		    if (conn [i].size >= 0)
		    {
			FD_SET (mdata [i],&wmask);
			conn [i].ready = TRUE;
			if (mdata [i] > maxfd)
			    maxfd = mdata [i];
			alldone = FALSE;
		    }
		    else
			conn [i].ready = FALSE;
		if (alldone && noread)
		    break;

 		j = select (maxfd + 1,0,&wmask,0,0);
		blocked = 0;
		while (blocked < j && j > 0)
		{
		    for (i = 0; i < cnncts; i++)
			if (conn [i].ready && conn [i].size >= 0 &&
			    FD_ISSET (mdata [i],&wmask))
			{
			    if ((conn [i].ready = writecnn (instr,i,&noread,
							    &curoff))
				< 0)
				goto data_err;
			    if (conn [i].ready == 0)
				blocked++;
			}
		        else if (conn [i].size == 0 && noread)
			    conn [i].size = -1;
		}
	    }
	    reply(226, "Transfer complete.", curoff);
	    return;

	default:
	  transflag = 0;
	  reply(550, "Unimplemented TYPE %d in send_data", type);
	  return;
    }

data_err:
    transflag = 0;
    perror_reply(426, "Data connection");
    return;

file_err:
    transflag = 0;
    perror_reply(551, "Error on input file");
}

/****************************************************************************/
/*                                                                          */
/* Process a client's request for a file (using multiple data connections). */
/*                                                                          */
/****************************************************************************/

void mretrieve (cmd, name)
char *cmd, *name;
{
    FILE *fin, *dout [MAXCNNCTS];
    struct stat st;
    int (*closefunc) __P((FILE *));
    int i;

    if (cmd == 0) 
    {
	fin = fopen(name, "r"), closefunc = fclose;
	st.st_size = 0;
    } 
    else 
    {
	char line[BUFSIZ];
	
	(void) sprintf(line, cmd, name), name = line;
	fin = ftpd_popen(line, "r"), closefunc = ftpd_pclose;
	st.st_size = -1;
	st.st_blksize = BUFSIZ;
    }
    if (fin == NULL) 
    {
	if (errno != 0) 
	{
	    perror_reply(550, name);
	    if (cmd == 0) 
		LOGCMD("get", name);
	}
	return;
    }
    byte_count = -1;
    if (cmd == 0 && (fstat(fileno(fin), &st) < 0 || !S_ISREG(st.st_mode))) 
    {
	reply(550, "%s: not a plain file.", name);
	goto done;
    }
    if (restart_point) 
    {
	if (type == TYPE_A) 
	{
	    off_t i, n;
	    int c;

	    n = restart_point;
	    i = 0;
	    while (i++ < n) 
	    {
		if ((c=getc(fin)) == EOF) 
		{
		    perror_reply(550, name);
		    goto done;
		}
		if (c == '\n')
		    i++;
	    }
	} 
	else if (lseek(fileno(fin), restart_point, L_SET) < 0) 
	{
	    perror_reply(550, name);
	    goto done;
	}
    }

    for (i = 0; i < MAXCNNCTS; i++)
    {
	mdata [i] = -1;
	conn [i].size = 0;
    }

    for (i = 0; i < cnncts; i++)
    {
	mdata_source [i].sin_port = htons(ntohs(ctrl_addr.sin_port) - 1);
	dout [i] = mdataconn(name, (off_t)-1, "r",i);
	if (dout [i] == NULL)
	    goto done;
    }
    reply(150, "Opening %s mode data connection for '%s'.",
	  type == TYPE_A ? "ASCII" : "BINARY", name);

    for (i = 0; i < cnncts; i++)
    {
	if (setsockopt (mdata [i],SOL_SOCKET,SO_SNDBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    goto done;
	if (setsockopt (mdata [i],SOL_SOCKET,SO_RCVBUF,(char *)&sbufsize,
			sizeof (sbufsize)) < 0)
	    goto done;
    }
    msend_data (fin, dout, st.st_blksize);
    for (i = 0; i < cnncts; i++)
    {
 	(void) fclose(dout [i]);
	mdata [i] = -1;
    }
    data = -1;
    pdata = -1;
    current = 0;
done:
    if (cmd == 0)
	LOGBYTES("get", name, byte_count);
    (*closefunc)(fin);
}
